# copyright 2003-2018 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.rset"""

import pickle
from urllib.parse import urlsplit

from logilab.common.testlib import TestCase, unittest_main, mock_object
from rql import parse

from cubicweb import NoResultError, MultipleResultsError
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.rset import NotAnEntity, ResultSet, attr_desc_iterator


def pprelcachedict(d):
    res = {}
    for k, (rset, related) in d.items():
        res[k] = sorted(v.eid for v in related)
    return sorted(res.items())


class AttrDescIteratorTC(TestCase):
    """TestCase for cubicweb.rset.attr_desc_iterator"""

    def test_relations_description(self):
        """tests relations_description() function"""
        queries = {
            "Any U,L,M where U is CWUser, U login L, U mail M": [
                (1, "login", "subject"),
                (2, "mail", "subject"),
            ],
            "Any U,L,M where U is CWUser, L is Foo, U mail M": [(2, "mail", "subject")],
            "Any C,P where C is Company, C employs P": [(1, "employs", "subject")],
            "Any C,P where C is Company, P employed_by P": [],
            "Any C where C is Company, C employs P": [],
        }
        for rql, relations in queries.items():
            result = list(attr_desc_iterator(parse(rql).children[0], 0, 0))
            self.assertEqual((rql, result), (rql, relations))

    def test_relations_description_indexed(self):
        """tests relations_description() function"""
        queries = {
            "Any C,U,P,L,M where C is Company, C employs P, U is CWUser, U login L, U mail M": {
                0: [(2, "employs", "subject")],
                1: [(3, "login", "subject"), (4, "mail", "subject")],
            },
        }
        for rql, results in queries.items():
            for idx, relations in results.items():
                result = list(attr_desc_iterator(parse(rql).children[0], idx, idx))
                self.assertEqual(result, relations)

    def test_subquery_callfunc(self):
        rql = (
            "Any A,B,C,COUNT(D) GROUPBY A,B,C WITH A,B,C,D BEING "
            "(Any YEAR(CD), MONTH(CD), S, X WHERE X is CWUser, "
            "X creation_date CD, X in_state S)"
        )
        rqlst = parse(rql)
        select, col = rqlst.locate_subquery(2, "CWUser", None)
        result = list(attr_desc_iterator(select, col, 2))
        self.assertEqual(result, [])

    def test_subquery_callfunc_2(self):
        rql = (
            "Any X,S,L WHERE "
            "X in_state S WITH X, L BEING (Any X,MAX(L) GROUPBY X WHERE "
            "X is CWUser, T wf_info_for X, T creation_date L)"
        )
        rqlst = parse(rql)
        select, col = rqlst.locate_subquery(0, "CWUser", None)
        result = list(attr_desc_iterator(select, col, 0))
        self.assertEqual(result, [(1, "in_state", "subject")])


class ResultSetTC(CubicWebTC):
    def setUp(self):
        super().setUp()
        self.rset = ResultSet(
            [[12, "adim"], [13, "syt"]],
            "Any U,L where U is CWUser, U login L",
            description=[["CWUser", "String"], ["Bar", "String"]],
        )
        self.rset.req = mock_object(vreg=self.vreg, repo=self.repo)

    def compare_urls(self, url1, url2):
        info1 = urlsplit(url1)
        info2 = urlsplit(url2)
        self.assertEqual(info1[:3], info2[:3])
        if info1[3] != info2[3]:
            params1 = dict(pair.split("=") for pair in info1[3].split("&"))
            params2 = dict(pair.split("=") for pair in info1[3].split("&"))
            self.assertDictEqual(params1, params2)

    def test_pickle(self):
        del self.rset.req
        rs2 = pickle.loads(pickle.dumps(self.rset))
        self.assertEqual(self.rset.rows, rs2.rows)
        self.assertEqual(self.rset.rowcount, rs2.rowcount)
        self.assertEqual(self.rset.rql, rs2.rql)
        self.assertEqual(self.rset.description, rs2.description)

    def test_build_url(self):
        with self.admin_access.cnx() as cnx:
            baseurl = cnx.base_url()
            self.compare_urls(
                cnx.build_url("view", vid="foo", rql="yo"),
                f"{baseurl}view?vid=foo&rql=yo",
            )
            self.compare_urls(
                cnx.build_url("view", _restpath="task/title/go"),
                f"{baseurl}task/title/go",
            )
            # self.compare_urls(cnx.build_url('view', _restpath='/task/title/go'),
            #                   '%stask/title/go' % baseurl)
            # empty _restpath should not crash
            self.compare_urls(cnx.build_url("view", _restpath=""), baseurl)

    def test_build(self):
        """test basic build of a ResultSet"""
        rs = ResultSet(
            [1, 2, 3], "CWGroup X", description=["CWGroup", "CWGroup", "CWGroup"]
        )
        self.assertEqual(rs.rowcount, 3)
        self.assertEqual(rs.rows, [1, 2, 3])
        self.assertEqual(rs.description, ["CWGroup", "CWGroup", "CWGroup"])

    def test_limit(self):
        rs = ResultSet(
            [[12000, "adim"], [13000, "syt"], [14000, "nico"]],
            "Any U,L where U is CWUser, U login L",
            description=[["CWUser", "String"]] * 3,
        )
        with self.admin_access.cnx() as cnx:
            rs.req = cnx
            rs.vreg = self.vreg
            self.assertEqual(rs.limit(2).rows, [[12000, "adim"], [13000, "syt"]])
            rs2 = rs.limit(2, offset=1)
            self.assertEqual(rs2.rows, [[13000, "syt"], [14000, "nico"]])
            self.assertEqual(rs2.get_entity(0, 0).cw_row, 0)
            self.assertEqual(rs.limit(2, offset=2).rows, [[14000, "nico"]])
            self.assertEqual(rs.limit(2, offset=3).rows, [])

    def test_limit_2(self):
        with self.admin_access.repo_cnx() as cnx:
            rs = cnx.execute("Any E,U WHERE E is CWEType, E created_by U")
            # get entity on row 9. This will fill its created_by relation cache,
            # with cwuser on row 9 as well
            e1 = rs.get_entity(9, 0)  # noqa
            # get entity on row 10. This will fill its created_by relation cache,
            # with cwuser built on row 9
            e2 = rs.get_entity(10, 0)
            # limit result set from row 10
            rs.limit(1, 10, inplace=True)
            # get back eid
            e = rs.get_entity(0, 0)
            self.assertIs(e2, e)
            # rs.limit has properly removed cwuser for cnxuest cache, but it's
            # still referenced by e/e2 relation cache
            u = e.created_by[0]
            # now ensure this doesn't trigger IndexError because cwuser.cw_row is 9
            # while now rset has only one row
            u.cw_rset[u.cw_row]

    def test_filter(self):
        rs = ResultSet(
            [[12000, "adim"], [13000, "syt"], [14000, "nico"]],
            "Any U,L where U is CWUser, U login L",
            description=[["CWUser", "String"]] * 3,
        )
        with self.admin_access.cnx() as cnx:
            rs.req = cnx
            rs.vreg = cnx.vreg

            def test_filter(entity):
                return entity.login != "nico"

            rs2 = rs.filtered_rset(test_filter)
            self.assertEqual(len(rs2), 2)
            self.assertEqual([login for _, login in rs2], ["adim", "syt"])
            self.assertEqual(rs2.description, rs.description[1:])

    def test_transform(self):
        rs = ResultSet(
            [[12, "adim"], [13, "syt"], [14, "nico"]],
            "Any U,L where U is CWUser, U login L",
            description=[["CWUser", "String"]] * 3,
        )
        with self.admin_access.cnx() as cnx:
            rs.req = cnx

            def test_transform(row, desc):
                return row[1:], desc[1:]

            rs2 = rs.transformed_rset(test_transform)

            self.assertEqual(len(rs2), 3)
            self.assertEqual(list(rs2), [["adim"], ["syt"], ["nico"]])

    def test_sort(self):
        rs = ResultSet(
            [[12000, "adim"], [13000, "syt"], [14000, "nico"]],
            "Any U,L where U is CWUser, U login L",
            description=[["CWUser", "String"]] * 3,
        )
        with self.admin_access.cnx() as cnx:
            rs.req = cnx
            rs.vreg = self.vreg

            rs2 = rs.sorted_rset(lambda e: e.cw_attr_cache["login"])
            self.assertEqual(len(rs2), 3)
            self.assertEqual([login for _, login in rs2], ["adim", "nico", "syt"])
            # make sure rs is unchanged
            self.assertEqual([login for _, login in rs], ["adim", "syt", "nico"])

            rs2 = rs.sorted_rset(lambda e: e.cw_attr_cache["login"], reverse=True)
            self.assertEqual(len(rs2), 3)
            self.assertEqual([login for _, login in rs2], ["syt", "nico", "adim"])
            # make sure rs is unchanged
            self.assertEqual([login for _, login in rs], ["adim", "syt", "nico"])

            rs3 = rs.sorted_rset(lambda row: row[1], col=-1)
            self.assertEqual(len(rs3), 3)
            self.assertEqual([login for _, login in rs3], ["adim", "nico", "syt"])
            # make sure rs is unchanged
            self.assertEqual([login for _, login in rs], ["adim", "syt", "nico"])

    def test_split(self):
        rs = ResultSet(
            [
                [12000, "adim", "Adim chez les pinguins"],
                [12000, "adim", "Jardiner facile"],
                [13000, "syt", "Le carrelage en 42 leçons"],
                [14000, "nico", "La tarte tatin en 15 minutes"],
                [14000, "nico", "L'épluchage du castor commun"],
            ],
            ("Any U, L, T WHERE U is CWUser, U login L," "D created_by U, D title T"),
            description=[["CWUser", "String", "String"]] * 5,
        )
        with self.admin_access.cnx() as cnx:
            rs.req = cnx
            rs.vreg = self.vreg
            rsets = rs.split_rset(lambda e: e.cw_attr_cache["login"])
            self.assertEqual(len(rsets), 3)
            self.assertEqual([login for _, login, _ in rsets[0]], ["adim", "adim"])
            self.assertEqual([login for _, login, _ in rsets[1]], ["syt"])
            self.assertEqual([login for _, login, _ in rsets[2]], ["nico", "nico"])
            # make sure rs is unchanged
            self.assertEqual(
                [login for _, login, _ in rs], ["adim", "adim", "syt", "nico", "nico"]
            )

            rsets = rs.split_rset(lambda e: e.cw_attr_cache["login"], return_dict=True)
            self.assertEqual(len(rsets), 3)
            self.assertEqual([login for _, login, _ in rsets["nico"]], ["nico", "nico"])
            self.assertEqual([login for _, login, _ in rsets["adim"]], ["adim", "adim"])
            self.assertEqual([login for _, login, _ in rsets["syt"]], ["syt"])
            # make sure rs is unchanged
            self.assertEqual(
                [login for _, login, _ in rs], ["adim", "adim", "syt", "nico", "nico"]
            )

            rsets = rs.split_rset(lambda s: s.count("d"), col=2)
            self.assertEqual(len(rsets), 2)
            self.assertEqual(
                [title for _, _, title in rsets[0]],
                [
                    "Adim chez les pinguins",
                    "Jardiner facile",
                    "L'épluchage du castor commun",
                ],
            )
            self.assertEqual(
                [title for _, _, title in rsets[1]],
                ["Le carrelage en 42 leçons", "La tarte tatin en 15 minutes"],
            )
            # make sure rs is unchanged
            self.assertEqual(
                [title for _, _, title in rs],
                [
                    "Adim chez les pinguins",
                    "Jardiner facile",
                    "Le carrelage en 42 leçons",
                    "La tarte tatin en 15 minutes",
                    "L'épluchage du castor commun",
                ],
            )

    def test_cached_syntax_tree(self):
        """make sure syntax tree is cached"""
        rqlst1 = self.rset.syntax_tree()
        rqlst2 = self.rset.syntax_tree()
        self.assertIs(rqlst1, rqlst2)

    def test_get_entity_simple(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(
                "CWUser",
                login="adim",
                upassword="adim",
                surname="di mascio",
                firstname="adrien",
            )
            cnx.drop_entity_cache()
            e = cnx.execute('Any X,T WHERE X login "adim", X surname T').get_entity(
                0, 0
            )
            self.assertEqual(e.cw_attr_cache["surname"], "di mascio")
            self.assertRaises(KeyError, e.cw_attr_cache.__getitem__, "firstname")
            self.assertRaises(KeyError, e.cw_attr_cache.__getitem__, "creation_date")
            self.assertEqual(pprelcachedict(e._cw_related_cache), [])
            e.complete()
            self.assertEqual(e.cw_attr_cache["firstname"], "adrien")
            self.assertEqual(pprelcachedict(e._cw_related_cache), [])

    def test_get_entity_advanced(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity("Bookmark", title="zou", path="/view")
            cnx.drop_entity_cache()
            cnx.execute('SET X bookmarked_by Y WHERE X is Bookmark, Y login "anon"')
            rset = cnx.execute(
                "Any X,Y,XT,YN WHERE X bookmarked_by Y, X title XT, Y login YN"
            )

            e = rset.get_entity(0, 0)
            self.assertEqual(e.cw_row, 0)
            self.assertEqual(e.cw_col, 0)
            self.assertEqual(e.cw_attr_cache["title"], "zou")
            self.assertRaises(KeyError, e.cw_attr_cache.__getitem__, "path")
            other_rset = cnx.execute("Any X, P WHERE X is Bookmark, X path P")
            # check that get_entity fetches e from the request's cache, and
            # updates it with attributes from the new rset
            self.assertIs(other_rset.get_entity(0, 0), e)
            self.assertIn("path", e.cw_attr_cache)
            self.assertEqual(pprelcachedict(e._cw_related_cache), [])

            e = rset.get_entity(0, 1)
            self.assertEqual(e.cw_row, 0)
            self.assertEqual(e.cw_col, 1)
            self.assertEqual(e.cw_attr_cache["login"], "anon")
            self.assertRaises(KeyError, e.cw_attr_cache.__getitem__, "firstname")
            self.assertEqual(pprelcachedict(e._cw_related_cache), [])
            e.complete()
            self.assertEqual(e.cw_attr_cache["firstname"], None)
            self.assertEqual(pprelcachedict(e._cw_related_cache), [])

            self.assertRaises(NotAnEntity, rset.get_entity, 0, 2)
            self.assertRaises(NotAnEntity, rset.get_entity, 0, 3)

    def test_get_entity_relation_cache_compt(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute('Any X,S WHERE X in_state S, X login "anon"')
            e = rset.get_entity(0, 0)
            seid = cnx.execute('State X WHERE X name "activated"')[0][0]
            # for_user / in_group are prefetched in CWUser __init__, in_state should
            # be filed from our query rset
            self.assertEqual(
                pprelcachedict(e._cw_related_cache), [("in_state_subject", [seid])]
            )

    def test_get_entity_advanced_prefilled_cache(self):
        with self.admin_access.repo_cnx() as cnx:
            e = cnx.create_entity("Bookmark", title="zou", path="path")
            cnx.commit()
            rset = cnx.execute(
                "Any X, U, S, XT, UL, SN WHERE X created_by U, U in_state S, "
                "X title XT, S name SN, U login UL, X eid %s" % e.eid
            )
            e = rset.get_entity(0, 0)
            self.assertEqual(e.cw_attr_cache["title"], "zou")
            self.assertEqual(
                pprelcachedict(e._cw_related_cache),
                [("created_by_subject", [cnx.user.eid])],
            )
            # first level of recursion
            u = e.created_by[0]
            self.assertEqual(u.cw_attr_cache["login"], "admin")
            self.assertRaises(KeyError, u.cw_attr_cache.__getitem__, "firstname")
            # second level of recursion
            s = u.in_state[0]
            self.assertEqual(s.cw_attr_cache["name"], "activated")
            self.assertRaises(KeyError, s.cw_attr_cache.__getitem__, "description")

    def test_get_entity_recursion(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity(
                "EmailAddress", address="toto", reverse_primary_email=cnx.user.eid
            )
            cnx.commit()

        # get_entity should fill the caches for user and email, even if both
        # entities are already in the connection's entity cache
        with self.admin_access.repo_cnx() as cnx:
            mail = cnx.find("EmailAddress").one()
            rset = cnx.execute("Any X, E WHERE X primary_email E")
            u = rset.get_entity(0, 0)
            self.assertTrue(u.cw_relation_cached("primary_email", "subject"))
            self.assertTrue(mail.cw_relation_cached("primary_email", "object"))

        with self.admin_access.repo_cnx() as cnx:
            mail = cnx.find("EmailAddress").one()
            rset = cnx.execute("Any X, E WHERE X primary_email E")
            rset.get_entity(0, 1)
            self.assertTrue(mail.cw_relation_cached("primary_email", "object"))
            u = cnx.user
            self.assertTrue(u.cw_relation_cached("primary_email", "subject"))

    def test_get_entity_cache_with_left_outer_join(self):
        with self.admin_access.cnx() as cnx:
            eid = cnx.execute(
                'INSERT CWUser E: E login "joe", E upassword "joe", E in_group G '
                'WHERE G name "users"'
            )[0][0]
            rset = cnx.execute(
                "Any X,E WHERE X eid %(x)s, X primary_email E?", {"x": eid}
            )
            e = rset.get_entity(0, 0)
            # if any of the assertion below fails with a KeyError, the relation is not cached
            # related entities should be an empty list
            self.assertEqual(e._cw_related_cache["primary_email_subject"][True], ())
            # related rset should be an empty rset
            cached = e._cw_related_cache["primary_email_subject"][False]
            self.assertIsInstance(cached, ResultSet)
            self.assertEqual(cached.rowcount, 0)

    def test_get_entity_union(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity("Bookmark", title="manger", path="path")
            cnx.drop_entity_cache()
            rset = cnx.execute(
                "Any X,N ORDERBY N WITH X,N BEING "
                "((Any X,N WHERE X is Bookmark, X title N)"
                " UNION "
                " (Any X,N WHERE X is CWGroup, X name N))"
            )
            expected = (
                ("CWGroup", "guests"),
                ("CWGroup", "managers"),
                ("Bookmark", "manger"),
                ("CWGroup", "owners"),
                ("CWGroup", "users"),
            )
            for entity in rset.entities():  # test get_entity for each row actually
                etype, n = expected[entity.cw_row]
                self.assertEqual(entity.cw_etype, etype)
                attr = etype == "Bookmark" and "title" or "name"
                self.assertEqual(entity.cw_attr_cache[attr], n)

    def test_one(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(
                "CWUser",
                login="cdevienne",
                upassword="cdevienne",
                surname="de Vienne",
                firstname="Christophe",
            )
            e = cnx.execute('Any X WHERE X login "cdevienne"').one()

            self.assertEqual(e.surname, "de Vienne")

            e = cnx.execute('Any X, N WHERE X login "cdevienne", X surname N').one()
            self.assertEqual(e.surname, "de Vienne")

            e = cnx.execute('Any N, X WHERE X login "cdevienne", X surname N').one(
                col=1
            )
            self.assertEqual(e.surname, "de Vienne")

    def test_one_no_rows(self):
        with self.admin_access.cnx() as cnx:
            with self.assertRaises(NoResultError):
                cnx.execute('Any X WHERE X login "patanok"').one()

    def test_one_multiple_rows(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(
                "CWUser",
                login="cdevienne",
                upassword="cdevienne",
                surname="de Vienne",
                firstname="Christophe",
            )

            cnx.create_entity(
                "CWUser",
                login="adim",
                upassword="adim",
                surname="di mascio",
                firstname="adrien",
            )

            with self.assertRaises(MultipleResultsError):
                cnx.execute("Any X WHERE X is CWUser").one()

    def test_first(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(
                "CWUser",
                login="cdevienne",
                upassword="cdevienne",
                surname="de Vienne",
                firstname="Christophe",
            )
            e = cnx.execute('Any X WHERE X login "cdevienne"').first()
            self.assertEqual(e.surname, "de Vienne")

            e = cnx.execute('Any X, N WHERE X login "cdevienne", X surname N').first()
            self.assertEqual(e.surname, "de Vienne")

            e = cnx.execute('Any N, X WHERE X login "cdevienne", X surname N').first(
                col=1
            )
            self.assertEqual(e.surname, "de Vienne")

    def test_first_no_rows(self):
        with self.admin_access.cnx() as cnx:
            with self.assertRaises(NoResultError):
                cnx.execute('Any X WHERE X login "patanok"').first()

    def test_first_multiple_rows(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(
                "CWUser",
                login="user1",
                upassword="cdevienne",
                surname="de Vienne",
                firstname="Christophe",
            )
            cnx.create_entity(
                "CWUser",
                login="user2",
                upassword="adim",
                surname="di mascio",
                firstname="adrien",
            )

            e = cnx.execute(
                "Any X ORDERBY X WHERE X is CWUser, " 'X login LIKE "user%"'
            ).first()
            self.assertEqual(e.login, "user1")

    def test_last(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(
                "CWUser",
                login="cdevienne",
                upassword="cdevienne",
                surname="de Vienne",
                firstname="Christophe",
            )
            e = cnx.execute('Any X WHERE X login "cdevienne"').last()
            self.assertEqual(e.surname, "de Vienne")

            e = cnx.execute('Any X, N WHERE X login "cdevienne", X surname N').last()
            self.assertEqual(e.surname, "de Vienne")

            e = cnx.execute('Any N, X WHERE X login "cdevienne", X surname N').last(
                col=1
            )
            self.assertEqual(e.surname, "de Vienne")

    def test_last_no_rows(self):
        with self.admin_access.cnx() as cnx:
            with self.assertRaises(NoResultError):
                cnx.execute('Any X WHERE X login "patanok"').last()

    def test_last_multiple_rows(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity(
                "CWUser",
                login="user1",
                upassword="cdevienne",
                surname="de Vienne",
                firstname="Christophe",
            )
            cnx.create_entity(
                "CWUser",
                login="user2",
                upassword="adim",
                surname="di mascio",
                firstname="adrien",
            )

            e = cnx.execute(
                "Any X ORDERBY X WHERE X is CWUser, " 'X login LIKE "user%"'
            ).last()
            self.assertEqual(e.login, "user2")

    def test_related_entity_optional(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity("Bookmark", title="aaaa", path="path")
            rset = cnx.execute("Any B,U,L WHERE B bookmarked_by U?, U login L")
            entity, rtype = rset.related_entity(0, 2)
            self.assertEqual(entity, None)
            self.assertEqual(rtype, None)

    def test_related_entity_union_subquery_1(self):
        with self.admin_access.cnx() as cnx:
            e = cnx.create_entity("Bookmark", title="aaaa", path="path")
            rset = cnx.execute(
                "Any X,N ORDERBY N WITH X,N BEING "
                "((Any X,N WHERE X is CWGroup, X name N)"
                " UNION "
                " (Any X,N WHERE X is Bookmark, X title N))"
            )
            entity, rtype = rset.related_entity(0, 1)
            self.assertEqual(entity.eid, e.eid)
            self.assertEqual(rtype, "title")
            self.assertEqual(entity.title, "aaaa")
            entity, rtype = rset.related_entity(1, 1)
            self.assertEqual(entity.cw_etype, "CWGroup")
            self.assertEqual(rtype, "name")
            self.assertEqual(entity.name, "guests")

    def test_related_entity_union_subquery_2(self):
        with self.admin_access.cnx() as cnx:
            e = cnx.create_entity("Bookmark", title="aaaa", path="path")
            rset = cnx.execute(
                "Any X,N ORDERBY N WHERE X is Bookmark WITH X,N BEING "
                "((Any X,N WHERE X is CWGroup, X name N)"
                " UNION "
                " (Any X,N WHERE X is Bookmark, X title N))"
            )
            entity, rtype = rset.related_entity(0, 1)
            self.assertEqual(entity.eid, e.eid)
            self.assertEqual(rtype, "title")
            self.assertEqual(entity.title, "aaaa")

    def test_related_entity_union_subquery_3(self):
        with self.admin_access.cnx() as cnx:
            e = cnx.create_entity("Bookmark", title="aaaa", path="path")
            rset = cnx.execute(
                "Any X,N ORDERBY N WITH N,X BEING "
                "((Any N,X WHERE X is CWGroup, X name N)"
                " UNION "
                " (Any N,X WHERE X is Bookmark, X title N))"
            )
            entity, rtype = rset.related_entity(0, 1)
            self.assertEqual(entity.eid, e.eid)
            self.assertEqual(rtype, "title")
            self.assertEqual(entity.title, "aaaa")

    def test_related_entity_union_subquery_4(self):
        with self.admin_access.cnx() as cnx:
            e = cnx.create_entity("Bookmark", title="aaaa", path="path")
            rset = cnx.execute(
                "Any X,X, N ORDERBY N WITH X,N BEING "
                "((Any X,N WHERE X is CWGroup, X name N)"
                " UNION "
                " (Any X,N WHERE X is Bookmark, X title N))"
            )
            entity, rtype = rset.related_entity(0, 2)
            self.assertEqual(entity.eid, e.eid)
            self.assertEqual(rtype, "title")
            self.assertEqual(entity.title, "aaaa")

    def test_related_entity_trap_subquery(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity("Bookmark", title="test bookmark", path="")
            cnx.execute('SET B bookmarked_by U WHERE U login "admin"')
            rset = cnx.execute(
                "Any B,T,L WHERE B bookmarked_by U, U login L "
                "WITH B,T BEING (Any B,T WHERE B is Bookmark, B title T)"
            )
            rset.related_entity(0, 2)

    def test_related_entity_subquery_outerjoin(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute(
                "Any X,S,L WHERE X in_state S "
                "WITH X, L BEING (Any X,MAX(L) GROUPBY X "
                "WHERE X is CWUser, T? wf_info_for X, T creation_date L)"
            )
            self.assertEqual(len(rset), 2)
            rset.related_entity(0, 1)
            rset.related_entity(0, 2)

    def test_entities(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute("Any U,G WHERE U in_group G")
            # make sure we have at least one element
            self.assertTrue(rset)
            self.assertEqual({e.e_schema.type for e in rset.entities(0)}, {"CWUser"})
            self.assertEqual({e.e_schema.type for e in rset.entities(1)}, {"CWGroup"})

    def test_iter_rows_with_entities(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute(
                "Any U,UN,G,GN WHERE U in_group G, U login UN, G name GN"
            )
            # make sure we have at least one element
            self.assertTrue(rset)
            out = list(rset.iter_rows_with_entities())[0]
            self.assertEqual(out[0].login, out[1])
            self.assertEqual(out[2].name, out[3])

    def test_printable_rql(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute("CWEType X WHERE X final FALSE")
            self.assertEqual(
                rset.printable_rql(), "Any X WHERE X final FALSE, X is CWEType"
            )

    def test_searched_text(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute('Any X WHERE X has_text "foobar"')
            self.assertEqual(rset.searched_text(), "foobar")
            rset = cnx.execute("Any X WHERE X has_text %(text)s", {"text": "foo"})
            self.assertEqual(rset.searched_text(), "foo")

    def test_count_users_by_date(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute(
                "Any D, COUNT(U) GROUPBY D " "WHERE U is CWUser, U creation_date D"
            )
            self.assertEqual(rset.related_entity(0, 0), (None, None))

    def test_str(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute("(Any X,N WHERE X is CWGroup, X name N)")
            self.assertIsInstance(str(rset), str)
            self.assertEqual(len(str(rset).splitlines()), 1)

    def test_repr(self):
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute("(Any X,N WHERE X is CWGroup, X name N)")
            self.assertIsInstance(repr(rset), str)
            self.assertGreater(len(repr(rset).splitlines()), 1)

            rset = cnx.execute('(Any X WHERE X is CWGroup, X name "managers")')
            self.assertIsInstance(str(rset), str)
            self.assertEqual(len(str(rset).splitlines()), 1)

    def test_slice(self):
        rs = ResultSet(
            [
                [12000, "adim", "Adim chez les pinguins"],
                [12000, "adim", "Jardiner facile"],
                [13000, "syt", "Le carrelage en 42 leçons"],
                [14000, "nico", "La tarte tatin en 15 minutes"],
                [14000, "nico", "L'épluchage du castor commun"],
            ],
            ("Any U, L, T WHERE U is CWUser, U login L," "D created_by U, D title T"),
            description=[["CWUser", "String", "String"]] * 5,
        )
        self.assertEqual(
            rs[1::2],
            [
                [12000, "adim", "Jardiner facile"],
                [14000, "nico", "La tarte tatin en 15 minutes"],
            ],
        )

    def test_nonregr_symmetric_relation(self):
        # see https://www.cubicweb.org/ticket/4739253
        with self.admin_access.client_cnx() as cnx:
            p1 = cnx.create_entity("Personne", nom="sylvain")
            cnx.create_entity("Personne", nom="denis", connait=p1)
            cnx.commit()
            rset = cnx.execute("Any X,Y WHERE X connait Y")
            rset.get_entity(0, 1)  # used to raise KeyError


if __name__ == "__main__":
    unittest_main()



#!/usr/bin/env python
"""
Parser for Javascript comments.
"""
import os.path as osp
import sys
import os
import getopt
import re


def clean_comment(match):
    comment = match.group()
    comment = strip_stars(comment)
    return comment


# Rest utilities
def rest_title(title, level, level_markups=["=", "=", "-", "~", "+", "`"]):
    size = len(title)
    if level == 0:
        return (
            "\n".join((level_markups[level] * size, title, level_markups[0] * size))
            + "\n"
        )
    return "\n".join(("\n" + title, level_markups[level] * size)) + "\n"


def get_doc_comments(text):
    """
    Return a list of all documentation comments in the file text.  Each
    comment is a pair, with the first element being the comment text and
    the second element being the line after it, which may be needed to
    guess function & arguments.

    >>> get_doc_comments(read_file('examples/module.js'))[0][0][:40]
    '/**\n * This is the module documentation.'
    >>> get_doc_comments(read_file('examples/module.js'))[1][0][7:50]
    'This is documentation for the first method.'
    >>> get_doc_comments(read_file('examples/module.js'))[1][1]
    'function the_first_function(arg1, arg2) '
    >>> get_doc_comments(read_file('examples/module.js'))[2][0]
    '/** This is the documentation for the second function. */'

    """
    return [
        clean_comment(match)
        for match in re.finditer(r"/\*\*.*?\*/", text, re.DOTALL | re.MULTILINE)
    ]


RE_STARS = re.compile(r"^\s*?\* ?", re.MULTILINE)


def strip_stars(doc_comment):
    """
    Strip leading stars from a doc comment.

    >>> strip_stars('/** This is a comment. */')
    'This is a comment.'
    >>> strip_stars('/**\n * This is a\n * multiline comment. */')
    'This is a\n multiline comment.'
    >>> strip_stars('/** \n\t * This is a\n\t * multiline comment. \n*/')
    'This is a\n multiline comment.'

    """
    return RE_STARS.sub("", doc_comment[3:-2]).strip()


def parse_js_files(args=sys.argv):
    """
    Main command-line invocation.
    """
    try:
        opts, args = getopt.gnu_getopt(
            args[1:], "p:o:h", ["jspath=", "output=", "help"]
        )
        opts = dict(opts)
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    rst_dir = opts.get("--output") or opts.get("-o")
    if rst_dir is None and len(args) != 1:
        rst_dir = "apidocs"
    js_dir = opts.get("--jspath") or opts.get("-p")
    if not osp.exists(osp.join(rst_dir)):
        os.makedirs(osp.join(rst_dir))

    index = set()
    for js_path, js_dirs, js_files in os.walk(js_dir):
        rst_path = re.sub(f"{js_dir}{osp.sep}*", "", js_path)
        for js_file in js_files:
            if not js_file.endswith(".js"):
                continue
            if js_file in FILES_TO_IGNORE:
                continue
            if not osp.exists(osp.join(rst_dir, rst_path)):
                os.makedirs(osp.join(rst_dir, rst_path))
            rst_content = extract_rest(js_path, js_file)
            filename = osp.join(rst_path, js_file[:-3])
            # add to index
            index.add(filename)
            # save rst file
            with open(osp.join(rst_dir, filename) + ".rst", "wb") as f_rst:
                f_rst.write(rst_content.encode())
    stream = open(osp.join(rst_dir, "index.rst"), "w")
    stream.write(
        """
Javascript API
==============

.. toctree::
    :maxdepth: 1

"""
    )
    # first write expected files in order
    for fileid in INDEX_IN_ORDER:
        try:
            index.remove(fileid)
        except Exception:
            raise Exception(
                "Bad file id %s referenced in INDEX_IN_ORDER in %s, "
                "fix this please" % (fileid, __file__)
            )
        stream.write("    %s\n" % fileid)
    # append remaining, by alphabetical order
    for fileid in sorted(index):
        stream.write("    %s\n" % fileid)
    stream.close()


def extract_rest(js_dir, js_file):
    js_filepath = osp.join(js_dir, js_file)
    filecontent = open(js_filepath, "U").read()
    comments = get_doc_comments(filecontent)
    rst = rest_title(js_file, 0)
    rst += ".. module:: %s\n\n" % js_file
    rst += "\n\n".join(comments)
    return rst


INDEX_IN_ORDER = [
    "cubicweb",
    "cubicweb.python",
    "cubicweb.htmlhelpers",
    "cubicweb.ajax",
    "cubicweb.ajax.box",
    "cubicweb.facets",
    "cubicweb.widgets",
    "cubicweb.image",
    "cubicweb.flot",
    "cubicweb.calendar",
    "cubicweb.preferences",
    "cubicweb.edition",
    "cubicweb.reledit",
]

FILES_TO_IGNORE = {
    "jquery.js",
    "jquery.treeview.js",
    "jquery.tablesorter.js",
    "jquery.timePicker.js",
    "jquery.flot.js",
    "jquery.corner.js",
    "jquery.ui.js",
    "excanvas.js",
    "gmap.utility.labeledmarker.js",
    "cubicweb.fckcwconfig.js",
    "cubicweb.fckcwconfig-full.js",
    "cubicweb.compat.js",
}

if __name__ == "__main__":
    parse_js_files()



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""
Provides a 'pyramid' command as a replacement to the 'start' command.

The reloading strategy is heavily inspired by (and partially copied from)
the pyramid script 'pserve'.
"""

import os
import signal
import subprocess
import sys
import threading
import time

import waitress
from logilab.common.configuration import merge_options
from logilab.common.shellutils import ASK

from cubicweb.cwctl import CWCTL, InstanceCommand, init_cmdline_log_threshold
from cubicweb.pyramid import wsgi_application_from_cwconfig
from cubicweb.pyramid.config import get_random_secret_key
from cubicweb.server import serverctl
from cubicweb.toolsutils import fill_templated_file, CommandHandler, underline_title

MAXFD = 1024


def _generate_pyramid_ini_file(pyramid_ini_path):
    """Write a 'pyramid.ini' file into apphome."""
    template_fpath = os.path.join(os.path.dirname(__file__), "pyramid.ini.tmpl")
    context = {
        "secret_1": get_random_secret_key(),
        "secret_2": get_random_secret_key(),
        "secret_3": get_random_secret_key(),
    }
    fill_templated_file(template_fpath, pyramid_ini_path, context)


class WebCreateHandler(CommandHandler):
    cmdname = "create"

    def bootstrap(self, cubes, automatic=False, inputlevel=0):
        """bootstrap this configuration"""
        if not automatic:
            print("\n" + underline_title("Generic web configuration"))
            config = self.config
            config.input_config("web", inputlevel)
            if ASK.confirm("Allow anonymous access ?", False):
                config.global_set_option("anonymous-user", "anon")
                config.global_set_option("anonymous-password", "anon")

    def postcreate(self, *args, **kwargs):
        """hooks called once instance's initialization has been completed"""


class AllInOneCreateHandler(serverctl.RepositoryCreateHandler, WebCreateHandler):
    """configuration to get an instance running in a Pyramid web server
    integrating a repository server in the same process
    """

    cfgname = "all-in-one"

    def bootstrap(self, cubes, automatic=False, inputlevel=0):
        """bootstrap this configuration"""
        serverctl.RepositoryCreateHandler.bootstrap(self, cubes, automatic, inputlevel)
        WebCreateHandler.bootstrap(self, cubes, automatic, inputlevel)
        _generate_pyramid_ini_file(os.path.join(self.config.apphome, "pyramid.ini"))


class PyramidStartHandler(InstanceCommand):
    """Start an interactive pyramid server.

    <instance>
      identifier of the instance to configure.
    """

    name = "start"
    actionverb = "started"

    options = merge_options(
        (
            (
                "debug-mode",
                {
                    "action": "store_true",
                    "help": (
                        "Activate the repository debug mode ("
                        "logs in the console and the debug toolbar)."
                    ),
                },
            ),
            (
                "debug",
                {
                    "short": "D",
                    "action": "store_true",
                    "help": 'Equals to "--debug-mode --reload"',
                },
            ),
            (
                "toolbar",
                {
                    "short": "t",
                    "action": "store_true",
                    "help": (
                        "Activate the pyramid debug toolbar"
                        '(the pypi "pyramid_debugtoolbar" package must be installed)'
                    ),
                },
            ),
            (
                "reload",
                {
                    "action": "store_true",
                    "help": "Restart the server if any source file is changed",
                },
            ),
            (
                "reload-interval",
                {
                    "type": "int",
                    "default": 1,
                    "help": "Interval, in seconds, between file modifications checks",
                },
            ),
            (
                "nb-threads",
                {
                    "type": "int",
                    "default": 4,
                    "help": "The number of threads used by the server.",
                },
            ),
        )
        + InstanceCommand.options
    )

    _reloader_environ_key = "CW_RELOADER_SHOULD_RUN"

    def debug(self, msg):
        print(f"DEBUG - {msg}")

    def info(self, msg):
        print(f"INFO - {msg}")

    def restart_with_reloader(self, filelist_path):
        self.debug("Starting subprocess with file monitor")

        # Create or clear monitored files list file.
        with open(filelist_path, "w") as f:
            pass

        while True:
            args = [sys.executable] + sys.argv
            new_environ = os.environ.copy()
            new_environ[self._reloader_environ_key] = "true"
            proc = None
            try:
                try:
                    proc = subprocess.Popen(args, env=new_environ)
                    exit_code = proc.wait()
                    proc = None
                    print("Process exited with", exit_code)
                except KeyboardInterrupt:
                    self.info("^C caught in monitor process")
                    return 1
            finally:
                if proc is not None:
                    proc.terminate()
                    self.info("Waiting for the server to stop. Hit CTRL-C to exit")
                    exit_code = proc.wait()

            if exit_code != 3:
                with open(filelist_path) as f:
                    filelist = [line.strip() for line in f]
                if filelist:
                    self.info("Reloading failed. Waiting for a file to change")
                    mon = Monitor(extra_files=filelist, nomodules=True)
                    while mon.check_reload():
                        time.sleep(1)
                else:
                    return exit_code

            self.info(f"{'-' * 20} Restarting {'-' * 20}")

    def set_needreload(self):
        self._needreload = True

    def install_reloader(self, poll_interval, extra_files, filelist_path):
        mon = Monitor(
            poll_interval=poll_interval,
            extra_files=extra_files,
            atexit=self.set_needreload,
            filelist_path=filelist_path,
        )
        mon_thread = threading.Thread(target=mon.periodic_reload)
        mon_thread.daemon = True
        mon_thread.start()

    def configfiles(self, cwconfig):
        """Generate instance configuration filenames"""
        yield cwconfig.main_config_file()
        for f in ("sources", "logging.conf", "pyramid.ini", "pyramid-debug.ini"):
            f = os.path.join(cwconfig.apphome, f)
            if os.path.exists(f):
                yield f

    def i18nfiles(self, cwconfig):
        """Generate instance i18n files"""
        i18ndir = os.path.join(cwconfig.apphome, "i18n")
        if os.path.exists(i18ndir):
            for lang in cwconfig.available_languages():
                f = os.path.join(i18ndir, lang, "LC_MESSAGES", "cubicweb.mo")
                if os.path.exists(f):
                    yield f

    def start_instance(self, appid):
        cwconfig = self.cwconfig

        cwconfig.in_debug_mode = self["debug"] or cwconfig["debug"]

        self._needreload = False
        autoreload = self["reload"] or cwconfig.in_debug_mode

        filelist_path = os.path.join(cwconfig.apphome, ".pyramid-reload-files.list")

        pyramid_ini_path = os.path.join(cwconfig.apphome, "pyramid.ini")
        if not os.path.exists(pyramid_ini_path):
            _generate_pyramid_ini_file(pyramid_ini_path)

        if autoreload and not os.environ.get(self._reloader_environ_key):
            return self.restart_with_reloader(filelist_path)

        if autoreload:
            _turn_sigterm_into_systemexit()
            self.debug("Running reloading file monitor")
            extra_files = [sys.argv[0]]
            extra_files.extend(self.configfiles(cwconfig))
            extra_files.extend(self.i18nfiles(cwconfig))
            self.install_reloader(
                self["reload-interval"], extra_files, filelist_path=filelist_path
            )

        # if no loglevel is specified and --debug is here, set log level at debug
        if self["loglevel"] is None and cwconfig.in_debug_mode:
            init_cmdline_log_threshold(self.cwconfig, "debug")

        # if the debugtoolbar is activated, test if it's importable
        if self["toolbar"]:
            try:
                import pyramid_debugtoolbar  # noqa
            except ImportError:
                print(
                    "Error: you've tried to activate the pyramid debugtoolbar but it failed to "
                    "import, make sure it's correctly installed by doing a "
                    "'pip install pyramid_debugtoolbar'.\nYou can find more information on the "
                    "official documentation: "
                    "https://docs.pylonsproject.org/projects/pyramid_debugtoolbar/en/latest/"
                )
                sys.exit(1)

        app = wsgi_application_from_cwconfig(
            cwconfig,
            debugtoolbar=self["toolbar"],
        )

        host = cwconfig["interface"]
        port = cwconfig["port"] or 8080
        url_scheme = "https" if cwconfig["base-url"].startswith("https") else "http"
        repo = app.application.registry["cubicweb.repository"]
        try:
            self.info(
                f"Application started. Point your browser to {cwconfig['base-url']}"
            )
            waitress.serve(
                app,
                host=host,
                port=port,
                url_scheme=url_scheme,
                clear_untrusted_proxy_headers=True,
                threads=self["nb-threads"],
            )
        finally:
            repo.shutdown()
        if self._needreload:
            return 3
        return 0


CWCTL.register(PyramidStartHandler)


def _turn_sigterm_into_systemexit():
    """Attempts to turn a SIGTERM exception into a SystemExit exception."""
    try:
        import signal
    except ImportError:
        return

    def handle_term(signo, frame):
        raise SystemExit

    signal.signal(signal.SIGTERM, handle_term)


class Monitor:
    """A file monitor and server stopper.

    It is a simplified version of pyramid pserve.Monitor, with little changes:

    -   The constructor takes extra_files, atexit, nomodules and filelist_path
    -   The process is stopped by auto-kill with signal SIGTERM
    """

    def __init__(
        self,
        poll_interval=1,
        extra_files=[],
        atexit=None,
        nomodules=False,
        filelist_path=None,
    ):
        self.module_mtimes = {}
        self.keep_running = True
        self.poll_interval = poll_interval
        self.extra_files = extra_files
        self.atexit = atexit
        self.nomodules = nomodules
        self.filelist_path = filelist_path

    def _exit(self):
        if self.atexit:
            self.atexit()
        os.kill(os.getpid(), signal.SIGTERM)

    def periodic_reload(self):
        while True:
            if not self.check_reload():
                self._exit()
                break
            time.sleep(self.poll_interval)

    def check_reload(self):
        filenames = list(self.extra_files)

        if not self.nomodules:
            for module in list(sys.modules.values()):
                try:
                    filename = module.__file__
                except (AttributeError, ImportError):
                    continue
                if filename is not None:
                    filenames.append(filename)

        for filename in filenames:
            try:
                stat = os.stat(filename)
                if stat:
                    mtime = stat.st_mtime
                else:
                    mtime = 0
            except OSError:
                continue
            if filename.endswith(".pyc") and os.path.exists(filename[:-1]):
                mtime = max(os.stat(filename[:-1]).st_mtime, mtime)
            if filename not in self.module_mtimes:
                self.module_mtimes[filename] = mtime
            elif self.module_mtimes[filename] < mtime:
                print(f"{filename} changed; reloading...")
                return False

        if self.filelist_path:
            with open(self.filelist_path) as f:
                filelist = {line.strip() for line in f}

            filelist.update(filenames)

            with open(self.filelist_path, "w") as f:
                for filename in filelist:
                    f.write(f"{filename}\n")

        return True



# copyright 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
from cubicweb import toolsutils
from cubicweb.devtools import DEFAULT_SOURCES
from cubicweb.devtools.apptest_config import ApptestConfiguration


class RealDatabaseConfiguration(ApptestConfiguration):
    init_repository = False
    sourcesdef = DEFAULT_SOURCES.copy()

    def sources(self):
        """
        By default, we run tests with the sqlite DB backend.
        One may use its own configuration by just creating a
        'sources' file in the test directory from wich tests are
        launched.
        """
        self._sources = self.sourcesdef
        return self._sources


def buildconfig(dbuser, dbpassword, dbname, adminuser, adminpassword, dbhost=None):
    """convenience function that builds a real-db configuration class"""
    sourcesdef = {
        "system": {
            "adapter": "native",
            "db-encoding": "UTF-8",  # 'ISO-8859-1',
            "db-user": dbuser,
            "db-password": dbpassword,
            "db-name": dbname,
            "db-driver": "postgres",
            "db-host": dbhost,
        },
        "admin": {
            "login": adminuser,
            "password": adminpassword,
        },
    }
    return type(
        "MyRealDBConfig", (RealDatabaseConfiguration,), {"sourcesdef": sourcesdef}
    )


def loadconfig(filename):
    """convenience function that builds a real-db configuration class
    from a file
    """
    return type(
        "MyRealDBConfig",
        (RealDatabaseConfiguration,),
        {"sourcesdef": toolsutils.read_config(filename)},
    )



# pylint: disable=W0622
"""cubicweb i18n test cube application packaging information"""

modname = "i18ntestcube"
distname = "cubicweb-i18ntestcube"

numversion = (0, 1, 0)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
description = "forum"
web = f"http://www.cubicweb.org/project/{distname}"

__depends__ = {
    "cubicweb": ">= 3.16.4",
}
__recommends__ = {}



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server import session


class HooksControlTC(CubicWebTC):
    def test_hooks_control(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(cnx._hooks_mode, session.HOOKS_ALLOW_ALL)
            self.assertEqual(cnx._hooks_categories, set())

            with cnx.deny_all_hooks_but("metadata"):
                self.assertEqual(cnx._hooks_mode, session.HOOKS_DENY_ALL)
                self.assertEqual(cnx._hooks_categories, {"metadata"})

                with cnx.deny_all_hooks_but():
                    self.assertEqual(cnx._hooks_categories, set())
                self.assertEqual(cnx._hooks_categories, {"metadata"})

                with cnx.deny_all_hooks_but("integrity"):
                    self.assertEqual(cnx._hooks_categories, {"integrity"})
                self.assertEqual(cnx._hooks_categories, {"metadata"})

                with cnx.allow_all_hooks_but("integrity"):
                    self.assertEqual(cnx._hooks_mode, session.HOOKS_ALLOW_ALL)
                    self.assertEqual(cnx._hooks_categories, {"integrity"})
                self.assertEqual(cnx._hooks_mode, session.HOOKS_DENY_ALL)
                self.assertEqual(cnx._hooks_categories, {"metadata"})

            self.assertEqual(cnx._hooks_mode, session.HOOKS_ALLOW_ALL)
            self.assertEqual(cnx._hooks_categories, set())


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""entity classes for optional library entities"""

from urllib.parse import urlsplit, urlunsplit

from logilab.mtconverter import xml_escape

from cubicweb import UnknownProperty
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.entity import _marker


def mangle_email(address):
    try:
        name, host = address.split("@", 1)
    except ValueError:
        return address
    return f"{name} at {host.replace('.', ' dot ')}"


class EmailAddress(AnyEntity):
    __regid__ = "EmailAddress"
    fetch_attrs, cw_fetch_order = fetch_config(["address", "alias"])
    rest_attr = "eid"

    def dc_title(self):
        if self.alias:
            return f"{self.alias} <{self.display_address()}>"
        return self.display_address()

    @property
    def email_of(self):
        return self.reverse_use_email and self.reverse_use_email[0] or None

    @property
    def prefered(self):
        return self.prefered_form and self.prefered_form[0] or self

    def related_emails(self, skipeids=None):
        # XXX move to eemail
        # check email relations are in the schema first
        subjrels = list(self.e_schema.object_relations.values())
        if not ("sender" in subjrels and "recipients" in subjrels):
            return
        rset = self._cw.execute(
            "DISTINCT Any X, S, D ORDERBY D DESC "
            "WHERE X sender Y or X recipients Y, "
            "X subject S, X date D, Y eid %(y)s",
            {"y": self.eid},
        )
        if skipeids is None:
            skipeids = set()
        for i in range(len(rset)):
            eid = rset[i][0]
            if eid in skipeids:
                continue
            skipeids.add(eid)
            yield rset.get_entity(i, 0)

    def display_address(self):
        if self._cw.vreg.config["mangle-emails"]:
            return mangle_email(self.address)
        return self.address

    def printable_value(self, attr, value=_marker, attrtype=None, format="text/html"):
        """overriden to return displayable address when necessary"""
        if attr == "address":
            address = self.display_address()
            if format == "text/html":
                address = xml_escape(address)
            return address
        return super().printable_value(attr, value, attrtype, format)


class Bookmark(AnyEntity):
    """customized class for Bookmark entities"""

    __regid__ = "Bookmark"
    fetch_attrs, cw_fetch_order = fetch_config(["title", "path"])

    def actual_url(self):
        url = self._cw.build_url(self.path)
        if self.title:
            urlparts = list(urlsplit(url))
            if urlparts[3]:
                urlparts[3] += f"&vtitle={self._cw.url_quote(self.title)}"
            else:
                urlparts[3] = f"vtitle={self._cw.url_quote(self.title)}"
            url = urlunsplit(urlparts)
        return url

    def action_url(self):
        return self.absolute_url() + "/follow"


class CWProperty(AnyEntity):
    __regid__ = "CWProperty"

    fetch_attrs, cw_fetch_order = fetch_config(["pkey", "value"])
    rest_attr = "pkey"

    def typed_value(self):
        return self._cw.vreg.typed_value(self.pkey, self.value)

    def dc_description(self, format="text/plain"):
        try:
            return self._cw._(self._cw.vreg.property_info(self.pkey)["help"])
        except UnknownProperty:
            return ""



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# copyright 2014-2016 UNLISH S.A.S. (Montpellier, FRANCE), all rights reserved.
#
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

"""Pyramid resources definitions for CubicWeb."""

from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPNotFound
from rql import TypeResolverException

from cubicweb import rdf, Unauthorized


def negociate_mime_type(request, possible_mimetypes):
    accepted_headers_by_weight = sorted(
        request.accept.parsed or [], key=lambda h: h[1], reverse=True
    )
    mime_type_negociated = None
    for parsed_header in accepted_headers_by_weight:
        accepted_mime_type = parsed_header[0]
        if accepted_mime_type in possible_mimetypes:
            mime_type_negociated = accepted_mime_type
            break
    return mime_type_negociated


def _get_entity_from_eid(request):
    try:
        entity = EntityResource.from_eid()(request).rset.one()
    except Unauthorized:
        raise HTTPNotFound()
    return entity


def _get_entity_from_identifier(request):
    try:
        entity = ETypeResource.from_match("etype")(request)[
            request.matchdict["identifier"]
        ].rset.one()
    except Unauthorized:
        raise HTTPNotFound()
    return entity


def download_context_from_eid(request):
    entity = _get_entity_from_eid(request)
    return BinaryResource(entity)


def download_context_from_identifier(request):
    entity = _get_entity_from_identifier(request)
    return BinaryResource(entity)


def rdf_context_from_eid(request):
    mime_type = negociate_mime_type(request, rdf.RDF_MIMETYPE_TO_FORMAT)
    entity = _get_entity_from_eid(request)
    if mime_type is None:
        return HTMLResource(entity)
    return RDFResource(entity, mime_type)


def rdf_context_from_identifier(request):
    mime_type = negociate_mime_type(request, rdf.RDF_MIMETYPE_TO_FORMAT)
    entity = _get_entity_from_identifier(request)
    if mime_type is None:
        return HTMLResource(entity)
    return RDFResource(entity, mime_type)


class RDFResource:
    def __init__(self, entity, mime_type):
        self.entity = entity
        self.mime_type = mime_type


class HTMLResource:
    def __init__(self, entity):
        self.entity = entity


class BinaryResource:
    def __init__(self, entity):
        self.entity = entity


class EntityResource:
    """A resource class for an entity. It provides method to retrieve an entity
    by eid.
    """

    @classmethod
    def from_eid(cls):
        def factory(request):
            return cls(request, None, None, request.matchdict["eid"])

        return factory

    def __init__(self, request, cls, attrname, value):
        self.request = request
        self.cls = cls
        self.attrname = attrname
        self.value = value

    @reify
    def rset(self):
        if self.cls is None:
            return self.request.cw_cnx.execute(
                "Any X WHERE X eid %(x)s", {"x": int(self.value)}
            )
        st = self.cls.fetch_rqlst(self.request.cw_cnx.user, ordermethod=None)
        st.add_constant_restriction(
            st.get_variable("X"), self.attrname, "x", "Substitute"
        )
        if self.attrname == "eid":
            try:
                rset = self.request.cw_cnx.execute(
                    st.as_string(), {"x": int(self.value)}
                )
            except (ValueError, TypeResolverException):
                # conflicting eid/type
                raise HTTPNotFound()
        else:
            rset = self.request.cw_cnx.execute(st.as_string(), {"x": self.value})
        return rset


class ETypeResource:
    """A resource for etype."""

    @classmethod
    def from_match(cls, matchname):
        def factory(request):
            return cls(request, request.matchdict[matchname])

        return factory

    def __init__(self, request, etype):
        vreg = request.registry["cubicweb.registry"]

        self.request = request
        self.etype = vreg.case_insensitive_etypes[etype.lower()]
        self.cls = vreg["etypes"].etype_class(self.etype)

    def __getitem__(self, value):
        (rest_attr, needcheck) = self.cls.cw_rest_attr_info()
        resource = EntityResource(self.request, self.cls, rest_attr, value)
        rset = resource.rset
        if rset.rowcount and (not needcheck or rset.rowcount == 1):
            return resource
        raise KeyError(value)

    @reify
    def rset(self):
        rql = self.cls.fetch_rql(self.request.cw_cnx.user)
        rset = self.request.cw_cnx.execute(rql)
        return rset



from urllib.parse import urljoin

import webtest

from pyramid.config import Configurator
from cubicweb.devtools import BASE_URL
from cubicweb.devtools.testlib import CubicWebTC


ACCEPTED_ORIGINS = ["example.com"]


class CustomTestApp(webtest.TestApp):
    def _gen_request(self, verb, url, *args, **kwargs):
        if not url.startswith(BASE_URL) and not url.startswith(("http://", "https://")):
            url = urljoin(BASE_URL, url)

        return super()._gen_request(verb, url, *args, **kwargs)

    def get(self, url, *args, **kwargs):
        if not url.startswith(BASE_URL) and not url.startswith(("http://", "https://")):
            url = urljoin(BASE_URL, url)

        return super().get(url, *args, **kwargs)


class _BasePyramidCWTest(CubicWebTC):
    settings = {}

    @classmethod
    def init_config(cls, config):
        super().init_config(config)
        config.global_set_option("anonymous-user", "anon")

    def _generate_pyramid_config(self):
        settings = {
            "cubicweb.session.secret": "test",
        }
        settings.update(self.settings)
        pyramid_config = Configurator(settings=settings)

        pyramid_config.registry["cubicweb.repository"] = self.repo
        pyramid_config.registry.settings["pyramid.debug_routematch"] = True
        pyramid_config.include("cubicweb.pyramid")

        self.includeme(pyramid_config)
        self.pyr_registry = pyramid_config.registry

        return pyramid_config

    def login(self, user=None, password=None, **args):
        return self.webapp.login(user, password, additional_arguments=args)

    def logout(self):
        return self.webapp.logout()


class PyramidCWTest(_BasePyramidCWTest):
    def includeme(self, config):
        pass

    def build_webapp(self):
        self.webapp = CustomTestApp(
            self.pyramid_config.make_wsgi_app(),
            extra_environ={"wsgi.url_scheme": "https"},
        )

    def setUp(self):
        # Skip CubicWebTestTC setUp
        super().setUp()
        settings = {
            "cubicweb.session.secret": "test",
        }
        settings.update(self.settings)
        self.pyramid_config = Configurator(settings=settings)

        self.pyramid_config.registry["cubicweb.repository"] = self.repo
        self.pyramid_config.include("cubicweb.pyramid")

        self.includeme(self.pyramid_config)
        self.pyr_registry = self.pyramid_config.registry
        self.build_webapp()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for schema rql (de)serialization"""

from logilab.database import get_db_helper
from yams import register_base_type, unregister_base_type

from cubicweb import Binary
from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.devtools.testlib import BaseTestCase as TestCase, CubicWebTC
from cubicweb.schema import CubicWebSchemaLoader
from cubicweb.server.schemaserial import (
    updateeschema2rql,
    updaterschema2rql,
    rschema2rql,
    eschema2rql,
    rdef2rql,
    specialize2rql,
    _erperms2rql as erperms2rql,
)

schema = config = None


def setUpModule(*args):
    register_base_type("BabarTestType", ("jungle_speed",))
    helper = get_db_helper("sqlite")
    helper.TYPE_MAPPING["BabarTestType"] = "TEXT"
    helper.TYPE_CONVERTERS["BabarTestType"] = lambda x: f'"{x}"'

    global schema, config
    loader = CubicWebSchemaLoader()
    config = ApptestConfiguration("data-schemaserial", __file__)
    config.bootstrap_cubes()
    schema = loader.load(config)


def tearDownModule(*args):
    global schema, config
    schema = config = None

    unregister_base_type("BabarTestType")
    helper = get_db_helper("sqlite")
    helper.TYPE_MAPPING.pop("BabarTestType", None)
    helper.TYPE_CONVERTERS.pop("BabarTestType", None)


cstrtypemap = {
    "RQLConstraint": "RQLConstraint_eid",
    "SizeConstraint": "SizeConstraint_eid",
    "StaticVocabularyConstraint": "StaticVocabularyConstraint_eid",
    "FormatConstraint": "FormatConstraint_eid",
}


class Schema2RQLTC(TestCase):
    def test_eschema2rql1(self):
        self.assertListEqual(
            [
                (
                    "INSERT CWEType X: X description %(description)s,X final %(final)s,X name %(name)s",
                    {
                        "description": (
                            "define a final relation: "
                            "link a final relation type from a non final entity "
                            "to a final entity type. used to build the instance schema"
                        ),
                        "name": "CWAttribute",
                        "final": False,
                    },
                )
            ],
            list(eschema2rql(schema.entity_schema_for("CWAttribute"))),
        )

    def test_eschema2rql2(self):
        self.assertListEqual(
            [
                (
                    "INSERT CWEType X: X description %(description)s,X final %(final)s,X name %(name)s",
                    {"description": "", "final": True, "name": "String"},
                )
            ],
            list(eschema2rql(schema.entity_schema_for("String"))),
        )

    def test_eschema2rql_specialization(self):
        # x: None since eschema.eid are None
        self.assertListEqual(
            [
                (
                    "SET X specializes ET WHERE X eid %(x)s, ET eid %(et)s",
                    {"et": None, "x": None},
                ),
                (
                    "SET X specializes ET WHERE X eid %(x)s, ET eid %(et)s",
                    {"et": None, "x": None},
                ),
            ],
            sorted(specialize2rql(schema)),
        )

    def test_esche2rql_custom_type(self):
        expected = [
            (
                "INSERT CWEType X: X description %(description)s,X final %(final)s,"
                "X name %(name)s",
                {"description": "", "name": "BabarTestType", "final": True},
            )
        ]
        got = list(eschema2rql(schema.entity_schema_for("BabarTestType")))
        self.assertListEqual(expected, got)

    def test_rschema2rql1(self):
        self.assertListEqual(
            [
                (
                    "INSERT CWRType X: X description %(description)s,X final %(final)s,"
                    "X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,"
                    "X name %(name)s,X symmetric %(symmetric)s",
                    {
                        "description": (
                            "link a relation definition to its relation type"
                        ),
                        "symmetric": False,
                        "name": "relation_type",
                        "final": False,
                        "fulltext_container": None,
                        "inlined": True,
                    },
                ),
                (
                    "INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,"
                    "X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,"
                    "X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": "",
                        "composite": "object",
                        "cardinality": "1*",
                        "ordernum": 1,
                    },
                ),
                (
                    "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X "
                    "WHERE CT eid %(ct)s, EDEF eid %(x)s",
                    {
                        "x": None,
                        "ct": "RQLConstraint_eid",
                        "value": '{"expression": "O final TRUE", "mainvars": ["O"], "msg": null}',
                    },
                ),
                (
                    "INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,"
                    "X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,"
                    "X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": "",
                        "composite": "object",
                        "ordernum": 1,
                        "cardinality": "1*",
                    },
                ),
                (
                    "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X "
                    "WHERE CT eid %(ct)s, EDEF eid %(x)s",
                    {
                        "x": None,
                        "ct": "RQLConstraint_eid",
                        "value": '{"expression": "O final FALSE", "mainvars": ["O"], "msg": null}',
                    },
                ),
            ],
            list(rschema2rql(schema.relation_schema_for("relation_type"), cstrtypemap)),
        )

    def test_rschema2rql2(self):
        self.assertListEqual(
            [
                (
                    "INSERT CWRType X: X description %(description)s,X final %(final)s,"
                    "X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,"
                    "X name %(name)s,X symmetric %(symmetric)s",
                    {
                        "description": "",
                        "symmetric": False,
                        "name": "add_permission",
                        "final": False,
                        "fulltext_container": None,
                        "inlined": False,
                    },
                ),
                (
                    "INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,"
                    "X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,"
                    "X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": (
                            "groups allowed to add entities/relations of this type"
                        ),
                        "composite": None,
                        "ordernum": 9999,
                        "cardinality": "**",
                    },
                ),
                (
                    "INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,"
                    "X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,"
                    "X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": "rql expression allowing to add entities/relations of this type",
                        "composite": "subject",
                        "ordernum": 9999,
                        "cardinality": "*?",
                    },
                ),
                (
                    "INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,"
                    "X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,"
                    "X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": (
                            "groups allowed to add entities/relations of this type"
                        ),
                        "composite": None,
                        "ordernum": 9999,
                        "cardinality": "**",
                    },
                ),
                (
                    "INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,"
                    "X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,"
                    "X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": "rql expression allowing to add entities/relations of this type",
                        "composite": "subject",
                        "ordernum": 9999,
                        "cardinality": "*?",
                    },
                ),
                (
                    "INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,"
                    "X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,"
                    "X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "cardinality": "**",
                        "composite": None,
                        "description": (
                            "groups allowed to add entities/relations of this type"
                        ),
                        "oe": None,
                        "ordernum": 9999,
                        "rt": None,
                        "se": None,
                    },
                ),
                (
                    "INSERT CWRelation X: X cardinality %(cardinality)s,X composite %(composite)s,"
                    "X description %(description)s,X ordernum %(ordernum)s,X relation_type ER,"
                    "X from_entity SE,X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "cardinality": "*?",
                        "composite": "subject",
                        "description": "rql expression allowing to add entities/relations of this type",
                        "oe": None,
                        "ordernum": 9999,
                        "rt": None,
                        "se": None,
                    },
                ),
            ],
            list(
                rschema2rql(schema.relation_schema_for("add_permission"), cstrtypemap)
            ),
        )

    def test_rschema2rql3(self):
        self.assertListEqual(
            [
                (
                    "INSERT CWRType X: X description %(description)s,X final %(final)s,"
                    "X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,"
                    "X name %(name)s,X symmetric %(symmetric)s",
                    {
                        "description": "",
                        "symmetric": False,
                        "name": "cardinality",
                        "final": True,
                        "fulltext_container": None,
                        "inlined": False,
                    },
                ),
                (
                    "INSERT CWAttribute X: X cardinality %(cardinality)s,X defaultval %(defaultval)s,"
                    "X description %(description)s,X formula %(formula)s,X fulltextindexed %(fulltextindexed)s,"
                    "X indexed %(indexed)s,X internationalizable %(internationalizable)s,"
                    "X ordernum %(ordernum)s,X relation_type ER,X from_entity SE,"
                    "X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": "subject/object cardinality",
                        "internationalizable": True,
                        "fulltextindexed": False,
                        "ordernum": 5,
                        "defaultval": None,
                        "indexed": False,
                        "formula": None,
                        "cardinality": "?1",
                    },
                ),
                (
                    "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X "
                    "WHERE CT eid %(ct)s, EDEF eid %(x)s",
                    {
                        "x": None,
                        "ct": "SizeConstraint_eid",
                        "value": '{"max": 2, "min": null, "msg": null}',
                    },
                ),
                (
                    "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X "
                    "WHERE CT eid %(ct)s, EDEF eid %(x)s",
                    {
                        "x": None,
                        "ct": "StaticVocabularyConstraint_eid",
                        "value": '{"msg": null, "values": ["?1", "11"]}',
                    },
                ),
                (
                    "INSERT CWAttribute X: X cardinality %(cardinality)s,X defaultval %(defaultval)s,"
                    "X description %(description)s,X formula %(formula)s,X fulltextindexed %(fulltextindexed)s,"
                    "X indexed %(indexed)s,X internationalizable %(internationalizable)s,"
                    "X ordernum %(ordernum)s,X relation_type ER,X from_entity SE,X to_entity OE "
                    "WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": "subject/object cardinality",
                        "internationalizable": True,
                        "fulltextindexed": False,
                        "ordernum": 5,
                        "defaultval": None,
                        "indexed": False,
                        "formula": None,
                        "cardinality": "?1",
                    },
                ),
                (
                    "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X "
                    "WHERE CT eid %(ct)s, EDEF eid %(x)s",
                    {
                        "x": None,
                        "ct": "SizeConstraint_eid",
                        "value": '{"max": 2, "min": null, "msg": null}',
                    },
                ),
                (
                    "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X "
                    "WHERE CT eid %(ct)s, EDEF eid %(x)s",
                    {
                        "x": None,
                        "ct": "StaticVocabularyConstraint_eid",
                        "value": (
                            '{"msg": null, "values": ["?*", "1*", "+*", "**", "?+", "1+", "++", "*+", "?1", '
                            '"11", "+1", "*1", "??", "1?", "+?", "*?"]}'
                        ),
                    },
                ),
            ],
            list(rschema2rql(schema.relation_schema_for("cardinality"), cstrtypemap)),
        )

    def test_rschema2rql_custom_type(self):
        expected = [
            (
                "INSERT CWRType X: X description %(description)s,X final %(final)s,"
                "X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,"
                "X name %(name)s,X symmetric %(symmetric)s",
                {
                    "description": "",
                    "final": True,
                    "fulltext_container": None,
                    "inlined": False,
                    "name": "custom_field_of_jungle",
                    "symmetric": False,
                },
            ),
            (
                "INSERT CWAttribute X: X cardinality %(cardinality)s,"
                "X defaultval %(defaultval)s,X description %(description)s,"
                "X extra_props %(extra_props)s,X formula %(formula)s,X indexed %(indexed)s,"
                "X ordernum %(ordernum)s,X relation_type ER,X from_entity SE,"
                "X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                {
                    "cardinality": "?1",
                    "defaultval": None,
                    "description": "",
                    "extra_props": b'{"jungle_speed": 42}',
                    "formula": None,
                    "indexed": False,
                    "oe": None,
                    "ordernum": 4,
                    "rt": None,
                    "se": None,
                },
            ),
        ]

        got = list(
            rschema2rql(
                schema.relation_schema_for("custom_field_of_jungle"), cstrtypemap
            )
        )
        self.assertEqual(2, len(got))
        # this is a custom type attribute with an extra parameter
        self.assertIn("extra_props", got[1][1])
        # this extr
        extra_props = got[1][1]["extra_props"]
        self.assertIsInstance(extra_props, Binary)
        got[1][1]["extra_props"] = got[1][1]["extra_props"].getvalue()
        self.assertListEqual(expected, got)

    def test_rdef2rql(self):
        self.assertListEqual(
            [
                (
                    "INSERT CWAttribute X: X cardinality %(cardinality)s,X defaultval %(defaultval)s,"
                    "X description %(description)s,X formula %(formula)s,X fulltextindexed %(fulltextindexed)s,"
                    "X indexed %(indexed)s,X internationalizable %(internationalizable)s,"
                    "X ordernum %(ordernum)s,X relation_type ER,X from_entity SE,"
                    "X to_entity OE WHERE SE eid %(se)s,ER eid %(rt)s,OE eid %(oe)s",
                    {
                        "se": None,
                        "rt": None,
                        "oe": None,
                        "description": "",
                        "internationalizable": True,
                        "fulltextindexed": False,
                        "ordernum": 3,
                        "defaultval": Binary.zpickle("text/plain"),
                        "indexed": False,
                        "formula": None,
                        "cardinality": "?1",
                    },
                ),
                (
                    "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X "
                    "WHERE CT eid %(ct)s, EDEF eid %(x)s",
                    {
                        "x": None,
                        "value": (
                            '{"msg": null, "values": ["text/rest", "text/markdown", '
                            '"text/html", "text/plain"]}'
                        ),
                        "ct": "FormatConstraint_eid",
                    },
                ),
                (
                    "INSERT CWConstraint X: X value %(value)s, X cstrtype CT, EDEF constrained_by X "
                    "WHERE CT eid %(ct)s, EDEF eid %(x)s",
                    {
                        "x": None,
                        "value": '{"max": 50, "min": null, "msg": null}',
                        "ct": "SizeConstraint_eid",
                    },
                ),
            ],
            list(
                rdef2rql(
                    schema["description_format"].relation_definitions[
                        ("CWRType", "String")
                    ],
                    cstrtypemap,
                )
            ),
        )

    def test_updateeschema2rql1(self):
        self.assertListEqual(
            [
                (
                    "SET X description %(description)s,X final %(final)s,"
                    "X name %(name)s WHERE X eid %(x)s",
                    {
                        "description": (
                            "define a final relation: link a final relation type from"
                            " a non final entity to a final entity type. used to build the instance schema"
                        ),
                        "x": 1,
                        "final": False,
                        "name": "CWAttribute",
                    },
                )
            ],
            list(updateeschema2rql(schema.entity_schema_for("CWAttribute"), 1)),
        )

    def test_updateeschema2rql2(self):
        self.assertListEqual(
            [
                (
                    "SET X description %(description)s,X final %(final)s,"
                    "X name %(name)s WHERE X eid %(x)s",
                    {"description": "", "x": 1, "final": True, "name": "String"},
                )
            ],
            list(updateeschema2rql(schema.entity_schema_for("String"), 1)),
        )

    def test_updaterschema2rql1(self):
        self.assertListEqual(
            [
                (
                    "SET X description %(description)s,X final %(final)s,"
                    "X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,"
                    "X name %(name)s,X symmetric %(symmetric)s WHERE X eid %(x)s",
                    {
                        "x": 1,
                        "symmetric": False,
                        "description": (
                            "link a relation definition to its relation type"
                        ),
                        "final": False,
                        "fulltext_container": None,
                        "inlined": True,
                        "name": "relation_type",
                    },
                )
            ],
            list(updaterschema2rql(schema.relation_schema_for("relation_type"), 1)),
        )

    def test_updaterschema2rql2(self):
        expected = [
            (
                "SET X description %(description)s,X final %(final)s,"
                "X fulltext_container %(fulltext_container)s,X inlined %(inlined)s,"
                "X name %(name)s,X symmetric %(symmetric)s WHERE X eid %(x)s",
                {
                    "x": 1,
                    "symmetric": False,
                    "description": "",
                    "final": False,
                    "fulltext_container": None,
                    "inlined": False,
                    "name": "add_permission",
                },
            )
        ]
        for i, (rql, args) in enumerate(
            updaterschema2rql(schema.relation_schema_for("add_permission"), 1)
        ):
            with self.subTest(i=i):
                self.assertEqual((rql, args), expected[i])


class Perms2RQLTC(TestCase):
    GROUP_MAPPING = {
        "managers": 0,
        "users": 1,
        "guests": 2,
        "owners": 3,
    }

    def test_eperms2rql1(self):
        self.assertListEqual(
            [
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 1}),
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 2}),
                ("SET X add_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
                ("SET X update_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
                ("SET X delete_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
            ],
            [
                (rql, kwargs)
                for rql, kwargs in erperms2rql(
                    schema.entity_schema_for("CWEType"), self.GROUP_MAPPING
                )
            ],
        )

    def test_rperms2rql2(self):
        self.assertListEqual(
            [
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 1}),
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 2}),
                ("SET X add_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
                ("SET X delete_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
            ],
            [
                (rql, kwargs)
                for rql, kwargs in erperms2rql(
                    schema.relation_schema_for("read_permission").relation_definition(
                        "CWEType", "CWGroup"
                    ),
                    self.GROUP_MAPPING,
                )
            ],
        )

    def test_rperms2rql3(self):
        self.assertListEqual(
            [
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 1}),
                ("SET X read_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 2}),
                ("SET X add_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
                ("SET X update_permission Y WHERE Y eid %(g)s, X eid %(x)s", {"g": 0}),
            ],
            [
                (rql, kwargs)
                for rql, kwargs in erperms2rql(
                    schema.relation_schema_for("name").relation_definition(
                        "CWEType", "String"
                    ),
                    self.GROUP_MAPPING,
                )
            ],
        )


class ComputedAttributeAndRelationTC(CubicWebTC):
    appid = "data-cwep002"

    def test(self):
        # force to read schema from the database
        self.repo.set_schema(self.repo.deserialize_schema(), resetvreg=False)
        schema = self.repo.schema
        self.assertEqual(
            [("Company", "Person")], sorted(schema["has_employee"].relation_definitions)
        )
        self.assertEqual(
            schema["has_employee"]
            .relation_definition("Company", "Person")
            .permissions["read"],
            ("managers",),
        )
        self.assertEqual("O works_for S", schema["has_employee"].rule)
        self.assertEqual(
            [("Company", "Int")], sorted(schema["total_salary"].relation_definitions)
        )
        self.assertEqual(
            "Any SUM(SA) GROUPBY X WHERE P works_for X, P salary SA",
            schema["total_salary"].relation_definitions["Company", "Int"].formula,
        )


if __name__ == "__main__":
    from unittest import main

    main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb.server.sources.ldapfeed unit and functional tests

Those tests expect to have slapd, python-ldap3 and ldapscripts packages installed.
"""

import os
import shutil
import threading
import subprocess
import sys
import tempfile
import time
import unittest
from os.path import join
from urllib.parse import urljoin

from yams import ValidationError
from cubicweb import AuthenticationError
from cubicweb.devtools import BASE_URL
from cubicweb.devtools.httptest import get_available_port
from cubicweb.devtools.testlib import CubicWebTC

CONFIG_LDAPFEED = """
user-base-dn=ou=People,dc=cubicweb,dc=test
group-base-dn=ou=Group,dc=cubicweb,dc=test
user-attrs-map=uid=login,mail=email,userPassword=upassword
group-attrs-map=cn=name,memberUid=member
"""

URL = None


def create_slapd_configuration(cls):
    global URL

    slapddir = tempfile.mkdtemp("cw-unittest-ldap")
    config = cls.config
    slapdconf = join(config.apphome, "slapd.conf")
    confin = open(join(config.apphome, "slapd.conf.in")).read()
    confstream = open(slapdconf, "w")
    confstream.write(confin % {"apphome": config.apphome, "testdir": slapddir})
    confstream.close()

    # fill ldap server with some data
    ldiffile = join(config.apphome, "ldap_test.ldif")
    config.info("Initing ldap database")
    cmdline = ["/usr/sbin/slapadd", "-d", "-1", "-f", slapdconf, "-l", ldiffile, "-c"]
    slapproc = subprocess.Popen(
        cmdline,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    (slapd_stdout, _) = slapproc.communicate(timeout=30)

    if slapproc.returncode:
        raise EnvironmentError(
            'Cannot initialize slapd database with cmdline="%s" returncode="%s" '
            '(from directory "%s")\n\nLDAP output:\n%s'
            % (
                " ".join(cmdline),
                slapproc.returncode,
                os.getcwd(),
                slapd_stdout.decode() if slapd_stdout is not None else slapd_stdout,
            )
        )

    # ldapuri = 'ldapi://' + join(basedir, "ldapi").replace('/', '%2f')
    port = get_available_port(range(9000, 9100))
    host = f"localhost:{port}"
    ldapuri = f"ldap://{host}"
    cmdline = ["/usr/sbin/slapd", "-f", slapdconf, "-h", ldapuri, "-d", "-1"]
    config.info("Starting slapd:", " ".join(cmdline))
    cls.slapd_process = subprocess.Popen(
        cmdline, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    )
    time.sleep(0.2)

    # what is this weird function you are asking yourself?
    # well... good question and the reason is pretty horrific: in 2023, slapd, the openldap
    # server is ... FULLY SYNCHRONOUS
    # what does it implies your ask? Well it implies that if you want slapd to write its debug
    # output on stdout and ... you don't read it, slapd, a server, will TOTALLY BLOCK
    # yes, it will wait for your to read its print, otherwise it will block and
    # won't do anything not accept any new connection
    # so yes, if you want some debug output for slapd, a server, you need another thread, just
    # to read its stdout/stderr (merged here) constantly
    # in 2023
    def consume_stdout(slapd):
        slapd.output = slapd.stdout.read()

    cls.thread = threading.Thread(target=consume_stdout, args=(cls.slapd_process,))
    cls.thread.start()

    if cls.slapd_process.poll() is None:
        config.info("slapd started with pid %s", cls.slapd_process.pid)
    else:
        cls.thread.join()
        raise EnvironmentError(
            'Cannot start slapd with cmdline="%s" returncode="%s" (from directory "%s")\n\n'
            "LDAP ouput:\n%s"
            % (
                " ".join(cmdline),
                cls.slapd_process.returncode,
                os.getcwd(),
                cls.slapd_process.output,
            )
        )
    URL = f"ldap://{host}"
    return slapddir


def terminate_slapd(cls):
    config = cls.config
    if cls.slapd_process and cls.slapd_process.returncode is None:
        config.info("terminating slapd")
        if hasattr(cls.slapd_process, "terminate"):
            cls.slapd_process.terminate()
        else:
            import signal

            os.kill(cls.slapd_process.pid, signal.SIGTERM)
        stdout, stderr = cls.slapd_process.communicate()
        if cls.slapd_process.returncode:
            print(
                "slapd returned with status: %s\n\nLDAP output:\n\n%s"
                % (cls.slapd_process.returncode, cls.slapd_process.output),
                file=sys.stderr,
            )
        config.info("DONE")
    shutil.rmtree(cls._tmpdir, ignore_errors=True)
    cls.thread.join()


def ldapsource(cnx):
    return cnx.find("CWSource", type="ldapfeed").one()


def update_source_config(source, options):
    config = source.dictconfig
    config.update(options)
    source.cw_set(config="\n".join("%s=%s" % x for x in config.items()))


@unittest.skip("The LDAP source will be removed with CubicWeb 5")
class LDAPFeedTestBase(CubicWebTC):
    test_db_id = "ldap-feed"
    loglevel = "ERROR"

    @classmethod
    def setUpClass(cls):
        super(LDAPFeedTestBase, cls).setUpClass()
        cls.init_slapd()

    @classmethod
    def tearDownClass(cls):
        cls.terminate_slapd()

    @classmethod
    def init_slapd(cls):
        for path in ("/usr/sbin/slapd", "/usr/sbin/slapadd"):
            if not os.path.exists(path):
                raise unittest.SkipTest(
                    f"{path} not found, please install the slapd and "
                    f"ldap-utils debian packages"
                )

        path = "/usr/bin/ldapmodify"
        if not os.path.exists(path):
            raise unittest.SkipTest(
                f"{path} not found, please install the ldap-utils debian " f"pacakage"
            )

        from cubicweb.cwctl import init_cmdline_log_threshold

        init_cmdline_log_threshold(cls.config, cls.loglevel)
        cls._tmpdir = create_slapd_configuration(cls)

    @classmethod
    def terminate_slapd(cls):
        terminate_slapd(cls)

    @classmethod
    def pre_setup_database(cls, cnx, config):
        cnx.create_entity(
            "CWSource",
            name="ldap",
            type="ldapfeed",
            parser="ldapfeed",
            url=URL,
            config=CONFIG_LDAPFEED,
        )
        cnx.commit()
        return cls.pull(cnx)

    @staticmethod
    def pull(cnx):
        lfsource = cnx.repo.source_by_uri("ldap")
        stats = lfsource.pull_data(cnx, force=True, raise_on_error=True)
        cnx.commit()
        return stats

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.execute('DELETE Any E WHERE E cw_source S, S name "ldap"')
            cnx.execute(
                "SET S config %(conf)s, S url %(url)s "
                'WHERE S is CWSource, S name "ldap"',
                {"conf": CONFIG_LDAPFEED, "url": URL},
            )
            cnx.commit()
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)

    def add_ldap_entry(self, dn, mods):
        """
        add an LDAP entity
        """
        modcmd = [f"dn: {dn}", "changetype: add"]
        for key, values in mods.items():
            if isinstance(values, str):
                values = [values]
            for value in values:
                modcmd.append(f"{key}: {value}")
        self._ldapmodify(modcmd)

    def delete_ldap_entry(self, dn):
        """
        delete an LDAP entity
        """
        modcmd = [f"dn: {dn}", "changetype: delete"]
        self._ldapmodify(modcmd)

    def update_ldap_entry(self, dn, mods):
        """
        modify one or more attributes of an LDAP entity
        """
        modcmd = [f"dn: {dn}", "changetype: modify"]
        for (kind, key), values in mods.items():
            modcmd.append(f"{kind}: {key}")
            if isinstance(values, str):
                values = [values]
            for value in values:
                modcmd.append(f"{key}: {value}")
            modcmd.append("-")
        self._ldapmodify(modcmd)

    def _ldapmodify(self, modcmd):
        uri = self.repo.source_by_uri("ldap").urls[0]
        updatecmd = [
            "ldapmodify",
            "-H",
            uri,
            "-v",
            "-x",
            "-D",
            "cn=admin,dc=cubicweb,dc=test",
            "-w",
            "cw",
        ]
        PIPE = subprocess.PIPE
        p = subprocess.Popen(updatecmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        p.stdin.write("\n".join(modcmd).encode("ascii"))
        p.stdin.close()
        if p.wait():
            raise RuntimeError(
                "ldap update failed: %s" % ("\n".join(p.stderr.readlines()))
            )


class CheckWrongGroup(LDAPFeedTestBase):
    """
    A testcase for situations where the default group for CWUser
    created from LDAP is wrongly configured.
    """

    def test_wrong_group(self):
        with self.admin_access.repo_cnx() as cnx:
            source = ldapsource(cnx)
            # inject a bogus group here, along with at least a valid one
            options = {"user-default-group": "thisgroupdoesnotexists,users"}
            update_source_config(source, options)
            cnx.commit()
            # here we emitted an error log entry
            source.repo_source.pull_data(cnx, force=True, raise_on_error=True)
            cnx.commit()


class LDAPFeedUserTC(LDAPFeedTestBase):
    """
    A testcase for CWUser support in ldapfeed (basic tests and authentication).
    """

    def assertMetadata(self, entity):
        self.assertTrue(entity.creation_date)
        self.assertTrue(entity.modification_date)

    def test_authenticate(self):
        source = self.repo.source_by_uri("ldap")
        with self.admin_access.repo_cnx() as cnx:
            # ensure we won't be logged against
            self.assertRaises(
                AuthenticationError, source.authenticate, cnx, "toto", "toto"
            )
            self.assertRaises(
                AuthenticationError, source.authenticate, cnx, "syt", "toto"
            )
            self.assertTrue(source.authenticate(cnx, "syt", "syt"))

    def test_ldapfeed_insert_collision(self):
        """
        when a user computed login from ldapfeed collides with a CWUser
        login the user MUST not be inserted, and message MUST be present
        at error level regarding the collision for troubleshooting purpose.
        We also check that in case the CWUser is skipped, the entity EmailAddress
        is not modified.
        If EmailAddress are not modified, CWGroup are not.
        """
        with self.admin_access.cnx() as cnx:
            user = cnx.find("CWUser", login="syt").one()
            user.cw_set(cw_source=cnx.find("CWSource", name="system").one())
            with cnx.security_enabled(write=False):
                user.cw_set(cwuri=urljoin(BASE_URL, str(user.eid)))
                for mail in user.use_email:
                    mail.cw_set(address=mail.address[:-3] + ".net")
            cnx.commit()
            with self.assertLogs("cubicweb.appobject", level="ERROR") as cm:
                self.pull(cnx)
                self.assertEqual(
                    cm.output,
                    [
                        "ERROR:cubicweb.appobject:not synchronizing user syt."
                        " User already exist in source system"
                    ],
                )
            for mail in user.use_email:
                self.assertTrue(mail.address.endswith(".net"))

    def test_base(self):
        with self.admin_access.repo_cnx() as cnx:
            # check a known one
            rset = cnx.execute("CWUser X WHERE X login %(login)s", {"login": "syt"})
            e = rset.get_entity(0, 0)
            self.assertEqual(e.login, "syt")
            e.complete()
            self.assertMetadata(e)
            self.assertEqual(e.firstname, None)
            self.assertEqual(e.surname, None)
            self.assertIn("users", set(g.name for g in e.in_group))
            self.assertEqual(e.owned_by[0].login, "syt")
            self.assertEqual(e.created_by, ())
            addresses = [pe.address for pe in e.use_email]
            addresses.sort()
            self.assertEqual(
                ["sylvain.thenault@logilab.fr", "syt@logilab.fr"], addresses
            )
            self.assertIn(
                e.primary_email[0].address,
                ["sylvain.thenault@logilab.fr", "syt@logilab.fr"],
            )
            # email content should be indexed on the user
            rset = cnx.execute('CWUser X WHERE X has_text "thenault"')
            self.assertEqual(rset.rows, [[e.eid]])

    def test_copy_to_system_source(self):
        "make sure we can 'convert' an LDAP user into a system one"
        with self.admin_access.repo_cnx() as cnx:
            source = self.repo.source_by_uri("ldap")
            eid = cnx.execute("CWUser X WHERE X login %(login)s", {"login": "syt"})[0][
                0
            ]
            cnx.execute(
                'SET X cw_source S WHERE X eid %(x)s, S name "system"', {"x": eid}
            )
            cnx.commit()
            rset = cnx.execute("CWUser X WHERE X login %(login)s", {"login": "syt"})
            self.assertEqual(len(rset), 1)
            e = rset.get_entity(0, 0)
            self.assertEqual(e.eid, eid)
            self.assertEqual(e.cw_source[0].name, "system")
            self.assertTrue(e.creation_date)
            self.assertTrue(e.modification_date)
            source.pull_data(cnx, raise_on_error=True)
            rset = cnx.execute("CWUser X WHERE X login %(login)s", {"login": "syt"})
            self.assertEqual(len(rset), 1)
            self.assertTrue(
                self.repo.system_source.authenticate(cnx, "syt", password="syt")
            )
            # make sure the pull from ldap have not "reverted" user as a ldap-feed user
            # and that the password stored in the system source is not empty or so
            user = cnx.execute('CWUser U WHERE U login "syt"').get_entity(0, 0)
            user.cw_clear_all_caches()
            self.assertEqual(user.cw_source[0].name, "system")
            cu = cnx.system_sql(
                "SELECT cw_upassword FROM cw_cwuser WHERE cw_login='syt';"
            )
            pwd = cu.fetchall()[0][0]
            self.assertIsNotNone(pwd)
            self.assertTrue(str(pwd))

    def test_bad_config(self):
        with self.admin_access.cnx() as cnx:
            with self.assertRaises(ValidationError) as cm:
                cnx.create_entity(
                    "CWSource",
                    name="erroneous",
                    type="ldapfeed",
                    parser="ldapfeed",
                    url="ldap.com",
                    config=CONFIG_LDAPFEED,
                )
            self.assertIn("badly formatted url", str(cm.exception))
            cnx.rollback()

            with self.assertRaises(ValidationError) as cm:
                cnx.create_entity(
                    "CWSource",
                    name="erroneous",
                    type="ldapfeed",
                    parser="ldapfeed",
                    url="http://ldap.com",
                    config=CONFIG_LDAPFEED,
                )
            self.assertIn("unsupported protocol", str(cm.exception))
            cnx.rollback()

            with self.assertRaises(ValidationError) as cm:
                cnx.create_entity(
                    "CWSource",
                    name="erroneous",
                    type="ldapfeed",
                    parser="ldapfeed",
                    url="ldap://host1\nldap://host2",
                    config=CONFIG_LDAPFEED,
                )
            self.assertIn("can only have one url", str(cm.exception))
            cnx.rollback()


class LDAPGeneratePwdTC(LDAPFeedTestBase):
    """
    A testcase for password generation on CWUser when none is imported
    """

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            lfsource = cnx.repo.source_by_uri("ldap")
            del lfsource.user_attrs["userPassword"]
        super(LDAPGeneratePwdTC, self).setup_database()

    def test_no_password(self):
        with self.admin_access.repo_cnx() as cnx:
            cu = cnx.system_sql(
                "SELECT cw_upassword FROM cw_cwuser WHERE cw_login='syt';"
            )
            pwd = cu.fetchall()[0][0]
            self.assertTrue(pwd)


class LDAPFeedUserDeletionTC(LDAPFeedTestBase):
    """
    A testcase for situations where users are deleted from or
    unavailable in the LDAP database.
    """

    def test_a_filter_inactivate(self):
        """filtered out people should be deactivated, unable to authenticate"""
        with self.admin_access.repo_cnx() as cnx:
            source = ldapsource(cnx)
            # filter with adim's phone number
            options = {"user-filter": "(telephonenumber=109)"}
            update_source_config(source, options)
            cnx.commit()
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
            repo_source = self.repo.source_by_uri("ldap")
            self.assertRaises(
                AuthenticationError, repo_source.authenticate, cnx, "syt", "syt"
            )
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(
                cnx.execute(
                    'Any N WHERE U login "syt", ' "U in_state S, S name N"
                ).rows[0][0],
                "deactivated",
            )
            self.assertEqual(
                cnx.execute(
                    'Any N WHERE U login "adim", ' "U in_state S, S name N"
                ).rows[0][0],
                "activated",
            )
            # unfilter, syt should be activated again
            source = ldapsource(cnx)
            options = {"user-filter": ""}
            update_source_config(source, options)
            cnx.commit()
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(
                cnx.execute(
                    'Any N WHERE U login "syt", ' "U in_state S, S name N"
                ).rows[0][0],
                "activated",
            )
            self.assertEqual(
                cnx.execute(
                    'Any N WHERE U login "adim", ' "U in_state S, S name N"
                ).rows[0][0],
                "activated",
            )

    def test_delete(self):
        """delete syt, pull, check deactivation, repull,
        read syt, pull, check activation
        """
        self.delete_ldap_entry("uid=syt,ou=People,dc=cubicweb,dc=test")
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
            source = self.repo.source_by_uri("ldap")
            self.assertRaises(
                AuthenticationError, source.authenticate, cnx, "syt", "syt"
            )
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(
                cnx.execute(
                    'Any N WHERE U login "syt", ' "U in_state S, S name N"
                ).rows[0][0],
                "deactivated",
            )
        with self.repo.internal_cnx() as cnx:
            # check that it doesn't choke
            self.pull(cnx)
        # reinsert syt
        self.add_ldap_entry(
            "uid=syt,ou=People,dc=cubicweb,dc=test",
            {
                "objectClass": [
                    "OpenLDAPperson",
                    "posixAccount",
                    "top",
                    "shadowAccount",
                ],
                "cn": "Sylvain Thenault",
                "sn": "Thenault",
                "gidNumber": "1004",
                "uid": "syt",
                "homeDirectory": "/home/syt",
                "shadowFlag": "134538764",
                "uidNumber": "1004",
                "givenName": "Sylvain",
                "telephoneNumber": "106",
                "displayName": "sthenault",
                "gecos": "Sylvain Thenault",
                "mail": ["sylvain.thenault@logilab.fr", "syt@logilab.fr"],
                "userPassword": "syt",
            },
        )
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(
                cnx.execute(
                    'Any N WHERE U login "syt", ' "U in_state S, S name N"
                ).rows[0][0],
                "activated",
            )

    def test_reactivate_deleted(self):
        # test reactivating BY HAND the user isn't enough to
        # authenticate, as the native source refuse to authenticate
        # user from other sources
        repo_source = self.repo.source_by_uri("ldap")
        self.delete_ldap_entry("uid=syt,ou=People,dc=cubicweb,dc=test")
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
        with self.admin_access.repo_cnx() as cnx:
            # reactivate user (which source is still ldap-feed)
            user = cnx.execute('CWUser U WHERE U login "syt"').get_entity(0, 0)
            user.cw_adapt_to("IWorkflowable").fire_transition("activate")
            cnx.commit()
            self.assertRaises(
                AuthenticationError, repo_source.authenticate, cnx, "syt", "syt"
            )

            # ok now let's try to make it a system user
            cnx.execute(
                'SET X cw_source S WHERE X eid %(x)s, S name "system"', {"x": user.eid}
            )
            cnx.commit()
            # and that we can now authenticate again
            self.assertRaises(
                AuthenticationError, repo_source.authenticate, cnx, "syt", "toto"
            )
            self.assertTrue(self.repo.authenticate_user(cnx, "syt", password="syt"))


class LDAPFeedGroupTC(LDAPFeedTestBase):
    """
    A testcase for group support in ldapfeed.
    """

    def test_groups_exist(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute('CWGroup X WHERE X name "dir"')
            self.assertEqual(len(rset), 1)

            rset = cnx.execute('CWGroup X WHERE X cw_source S, S name "ldap"')
            self.assertEqual(len(rset), 2)

    def test_group_deleted(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute('CWGroup X WHERE X name "dir"')
            self.assertEqual(len(rset), 1)

    def test_in_group(self):
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("CWGroup X WHERE X name %(name)s", {"name": "dir"})
            dirgroup = rset.get_entity(0, 0)
            self.assertEqual(
                set(["syt", "adim"]), set([u.login for u in dirgroup.reverse_in_group])
            )
            rset = cnx.execute("CWGroup X WHERE X name %(name)s", {"name": "logilab"})
            logilabgroup = rset.get_entity(0, 0)
            self.assertEqual(
                set(["adim"]), set([u.login for u in logilabgroup.reverse_in_group])
            )

    def test_group_member_added(self):
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute(
                "Any L WHERE U in_group G, G name %(name)s, U login L",
                {"name": "logilab"},
            )
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][0], "adim")

        try:
            self.update_ldap_entry(
                "cn=logilab,ou=Group,dc=cubicweb,dc=test",
                {("add", "memberUid"): ["syt"]},
            )
            with self.repo.internal_cnx() as cnx:
                self.pull(cnx)

            with self.admin_access.repo_cnx() as cnx:
                rset = cnx.execute(
                    "Any L WHERE U in_group G, G name %(name)s, U login L",
                    {"name": "logilab"},
                )
                self.assertEqual(len(rset), 2)
                members = set([u[0] for u in rset])
                self.assertEqual(set(["adim", "syt"]), members)

        finally:
            # back to normal ldap setup
            self.terminate_slapd()
            self.init_slapd()

    def test_group_member_deleted(self):
        with self.repo.internal_cnx() as cnx:
            self.pull(cnx)  # ensure we are sync'ed
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute(
                "Any L WHERE U in_group G, G name %(name)s, U login L",
                {"name": "logilab"},
            )
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset[0][0], "adim")

        try:
            self.update_ldap_entry(
                "cn=logilab,ou=Group,dc=cubicweb,dc=test",
                {("delete", "memberUid"): ["adim"]},
            )
            with self.repo.internal_cnx() as cnx:
                self.pull(cnx)

            with self.admin_access.repo_cnx() as cnx:
                rset = cnx.execute(
                    "Any L WHERE U in_group G, G name %(name)s, U login L",
                    {"name": "logilab"},
                )
                self.assertEqual(len(rset), 0, rset.rows)
        finally:
            # back to normal ldap setup
            self.terminate_slapd()
            self.init_slapd()


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()
